@@ -1,6 +1,6 @@
 /*
  * LibSylph -- E-Mail client library
- * Copyright (C) 1999-2007 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2010 Hiroyuki Yamamoto
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -43,11 +43,22 @@
 #include "prefs_account.h"
 #include "sylmain.h"
 
+typedef struct _FolderPrivData FolderPrivData;
+
+struct _FolderPrivData {
+	Folder *folder;
+	FolderItem *junk;
+	gpointer data;
+};
+
 static GList *folder_list = NULL;
+static GList *folder_priv_list = NULL;
 
 static void folder_init		(Folder		*folder,
 				 const gchar	*name);
 
+static FolderPrivData *folder_get_priv	(Folder		*folder);
+
 static gboolean folder_read_folder_func	(GNode		*node,
 					 gpointer	 data);
 static gchar *folder_get_list_path	(void);
@@ -110,18 +121,28 @@ void folder_remote_folder_init(Folder *folder, const gchar *name,
 {
 	folder_init(folder, name);
 	REMOTE_FOLDER(folder)->session = NULL;
+	REMOTE_FOLDER(folder)->remove_cache_on_destroy = TRUE;
 }
 
 void folder_destroy(Folder *folder)
 {
+	FolderPrivData *priv;
+
 	g_return_if_fail(folder != NULL);
 	g_return_if_fail(folder->klass->destroy != NULL);
 
+	debug_print("folder_destroy: destroying Folder (%p)\n", folder);
+
 	folder->klass->destroy(folder);
 
 	folder_list = g_list_remove(folder_list, folder);
 
 	folder_tree_destroy(folder);
+
+	priv = folder_get_priv(folder);
+	folder_priv_list = g_list_remove(folder_priv_list, priv);
+	g_free(priv);
+
 	g_free(folder->name);
 	g_free(folder);
 }
@@ -141,6 +162,26 @@ void folder_remote_folder_destroy(RemoteFolder *rfolder)
 		session_destroy(rfolder->session);
 }
 
+gint folder_remote_folder_destroy_all_sessions(void)
+{
+	GList *list;
+	Folder *folder;
+	RemoteFolder *rfolder;
+
+	for (list = folder_list; list != NULL; list = list->next) {
+		folder = FOLDER(list->data);
+		if (FOLDER_IS_REMOTE(folder)) {
+			rfolder = REMOTE_FOLDER(folder);
+			if (rfolder->session) {
+				session_destroy(rfolder->session);
+				rfolder->session = NULL;
+			}
+		}
+	}
+
+	return 0;
+}
+
 gint folder_scan_tree(Folder *folder)
 {
 	g_return_val_if_fail(folder != NULL, -1);
@@ -311,6 +352,8 @@ void folder_item_destroy(FolderItem *item)
 			folder->queue = NULL;
 		else if (folder->trash == item)
 			folder->trash = NULL;
+		else if (folder_get_junk(folder) == item)
+			folder_set_junk(folder, NULL);
 	}
 
 	g_free(item->name);
@@ -397,6 +440,9 @@ void folder_add(Folder *folder)
 	Folder *cur_folder;
 	GList *cur;
 	gint i;
+	FolderPrivData *priv;
+
+	debug_print("Adding Folder (%p) to folder list\n", folder);
 
 	g_return_if_fail(folder != NULL);
 
@@ -415,6 +461,9 @@ void folder_add(Folder *folder)
 	}
 
 	folder_list = g_list_insert(folder_list, folder, i);
+	priv = g_new0(FolderPrivData, 1);
+	priv->folder = folder;
+	folder_priv_list = g_list_insert(folder_priv_list, priv, i);
 }
 
 GList *folder_get_list(void)
@@ -474,7 +523,8 @@ void folder_write_list(void)
 	if (prefs_file_close(pfile) < 0)
 		g_warning("failed to write folder list.\n");
 
-	g_signal_emit_by_name(syl_app_get(), "folderlist-updated");
+	if (syl_app_get())
+		g_signal_emit_by_name(syl_app_get(), "folderlist-updated");
 }
 
 struct TotalMsgStatus
@@ -647,10 +697,19 @@ FolderItem *folder_find_child_item_by_name(FolderItem *item, const gchar *name)
 {
 	GNode *node;
 	FolderItem *child;
+	const gchar *base;
+
+	if (!name)
+		return NULL;
 
 	for (node = item->node->children; node != NULL; node = node->next) {
 		child = FOLDER_ITEM(node->data);
-		if (strcmp2(g_basename(child->path), name) == 0)
+		base = g_basename(child->path);
+#ifdef G_OS_WIN32
+		if (base && g_ascii_strcasecmp(base, name) == 0)
+#else
+		if (strcmp2(base, name) == 0)
+#endif
 			return child;
 	}
 
@@ -847,6 +906,57 @@ FolderItem *folder_get_default_trash(void)
 	return folder->trash;
 }
 
+FolderItem *folder_get_default_junk(void)
+{
+	FolderPrivData *priv;
+
+	if (!folder_list) return NULL;
+	if (!folder_priv_list) return NULL;
+	priv = (FolderPrivData *)folder_priv_list->data;
+	g_return_val_if_fail(priv != NULL, NULL);
+	g_return_val_if_fail(priv->folder != NULL, NULL);
+	return priv->junk;
+}
+
+static FolderPrivData *folder_get_priv(Folder *folder)
+{
+	FolderPrivData *priv;
+	GList *cur;
+
+	g_return_val_if_fail(folder != NULL, NULL);
+
+	for (cur = folder_priv_list; cur != NULL; cur = cur->next) {
+		priv = (FolderPrivData *)cur->data;
+		if (priv->folder == folder)
+			return priv;
+	}
+
+	g_warning("folder_get_priv: private data for Folder (%p) not found.",
+		  folder);
+
+	return NULL;
+}
+
+FolderItem *folder_get_junk(Folder *folder)
+{
+	FolderPrivData *priv;
+
+	priv = folder_get_priv(folder);
+	if (priv)
+		return priv->junk;
+
+	return NULL;
+}
+
+void folder_set_junk(Folder *folder, FolderItem *item)
+{
+	FolderPrivData *priv;
+
+	priv = folder_get_priv(folder);
+	if (priv)
+		priv->junk = item;
+}
+
 #define CREATE_FOLDER_IF_NOT_EXIST(member, dir, type)	\
 {							\
 	if (!folder->member) {				\
@@ -871,7 +981,7 @@ void folder_set_missing_folders(void)
 		g_return_if_fail(rootitem != NULL);
 
 		if (folder->inbox && folder->outbox && folder->draft &&
-		    folder->queue && folder->trash)
+		    folder->queue && folder->trash && folder_get_junk(folder))
 			continue;
 
 		if (folder_create_tree(folder) < 0) {
@@ -885,7 +995,13 @@ void folder_set_missing_folders(void)
 		CREATE_FOLDER_IF_NOT_EXIST(draft,  DRAFT_DIR,  F_DRAFT);
 		CREATE_FOLDER_IF_NOT_EXIST(queue,  QUEUE_DIR,  F_QUEUE);
 		CREATE_FOLDER_IF_NOT_EXIST(trash,  TRASH_DIR,  F_TRASH);
-		/* CREATE_FOLDER_IF_NOT_EXIST(junk,   JUNK_DIR,  F_JUNK); */
+
+		if (!folder_get_junk(folder)) {
+			item = folder_item_new(JUNK_DIR, JUNK_DIR);
+			item->stype = F_JUNK;
+			folder_item_append(rootitem, item);
+			folder_set_junk(folder, item);
+		}
 	}
 }
 
@@ -1143,6 +1259,37 @@ gint folder_item_add_msgs(FolderItem *dest, GSList *file_list,
 				       first);
 }
 
+gint folder_item_add_msg_msginfo(FolderItem *dest, MsgInfo *msginfo,
+				 gboolean remove_source)
+{
+	Folder *folder;
+
+	g_return_val_if_fail(dest != NULL, -1);
+	g_return_val_if_fail(msginfo != NULL, -1);
+	g_return_val_if_fail(msginfo->file_path != NULL, -1);
+	g_return_val_if_fail(dest->folder->klass->add_msg_msginfo != NULL, -1);
+
+	folder = dest->folder;
+
+	return folder->klass->add_msg_msginfo(folder, dest, msginfo,
+					      remove_source);
+}
+
+gint folder_item_add_msgs_msginfo(FolderItem *dest, GSList *msglist,
+				  gboolean remove_source, gint *first)
+{
+	Folder *folder;
+
+	g_return_val_if_fail(dest != NULL, -1);
+	g_return_val_if_fail(msglist != NULL, -1);
+	g_return_val_if_fail(dest->folder->klass->add_msgs_msginfo != NULL, -1);
+
+	folder = dest->folder;
+
+	return folder->klass->add_msgs_msginfo(folder, dest, msglist,
+					       remove_source, first);
+}
+
 gint folder_item_move_msg(FolderItem *dest, MsgInfo *msginfo)
 {
 	Folder *folder;
@@ -1350,10 +1497,8 @@ static gboolean folder_build_tree(GNode *node, gpointer data)
 				stype = F_QUEUE;
 			else if (!g_ascii_strcasecmp(attr->value, "trash"))
 				stype = F_TRASH;
-#if 0
 			else if (!g_ascii_strcasecmp(attr->value, "junk"))
 				stype = F_JUNK;
-#endif
 			else if (!g_ascii_strcasecmp(attr->value, "virtual"))
 				stype = F_VIRTUAL;
 		} else if (!strcmp(attr->name, "name"))
@@ -1426,6 +1571,10 @@ static gboolean folder_build_tree(GNode *node, gpointer data)
 				qsearch_cond_type = 5;
 			else if (!strcmp(attr->value, "last5"))
 				qsearch_cond_type = 6;
+			else if (!strcmp(attr->value, "last7"))
+				qsearch_cond_type = 7;
+			else if (!strcmp(attr->value, "in-addressbook"))
+				qsearch_cond_type = 8;
 		} else if (!strcmp(attr->name, "account_id")) {
 			account = account_find_from_id(atoi(attr->value));
 			if (!account) g_warning("account_id: %s not found\n",
@@ -1473,7 +1622,7 @@ static gboolean folder_build_tree(GNode *node, gpointer data)
 	case F_DRAFT:  folder->draft  = item; break;
 	case F_QUEUE:  folder->queue  = item; break;
 	case F_TRASH:  folder->trash  = item; break;
-	/* case F_JUNK:   folder->junk   = item; break; */
+	case F_JUNK:   folder_set_junk(folder, item); break;
 	default:       break;
 	}
 	item->account = account;
@@ -1603,7 +1752,8 @@ static void folder_write_list_recursive(GNode *node, gpointer data)
 					"from", "subject", "score", "label",
 					"mark", "unread", "mime", "to"};
 	static gchar *qsearch_cond_str[] = {"all", "unread", "mark", "clabel",
-					    "mime", "w1day", "last5"};
+					    "mime", "w1day", "last5", "last7",
+					    "in-addressbook"};
 
 	g_return_if_fail(node != NULL);
 	g_return_if_fail(fp != NULL);
@@ -1662,7 +1812,7 @@ static void folder_write_list_recursive(GNode *node, gpointer data)
 				fprintf(fp, " sort_type=\"descending\"");
 		}
 		if (item->qsearch_cond_type > 0 &&
-		    item->qsearch_cond_type < 7) {
+		    item->qsearch_cond_type < 9) {
 			fprintf(fp, " qsearch_cond=\"%s\"",
 				qsearch_cond_str[item->qsearch_cond_type]);
 		}
