@@ -1,6 +1,6 @@
 /*
  * LibSylph -- E-Mail client library
- * Copyright (C) 1999-2007 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2010 Hiroyuki Yamamoto
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -42,6 +42,9 @@
 #include <time.h>
 
 #ifdef G_OS_WIN32
+#ifndef WINVER
+#  define WINVER 0x0500
+#endif
 #  include <windows.h>
 #  include <wchar.h>
 #  include <direct.h>
@@ -149,6 +152,65 @@ gint g_chmod(const gchar *path, gint mode)
 }
 #endif /* GLIB_CHECK_VERSION && G_OS_UNIX */
 
+#ifndef G_OS_UNIX
+gint syl_link(const gchar *src, const gchar *dest)
+{
+#ifdef G_OS_WIN32
+	wchar_t *wsrc;
+	wchar_t *wdest;
+	gint retval;
+	gint save_errno;
+
+	wsrc = g_utf8_to_utf16(src, -1, NULL, NULL, NULL);
+	if (wsrc == NULL) {
+		errno = EINVAL;
+		return -1;
+	}
+	wdest = g_utf8_to_utf16(dest, -1, NULL, NULL, NULL);
+	if (wdest == NULL) {
+		g_free(wsrc);
+		errno = EINVAL;
+		return -1;
+	}
+
+	errno = 0;
+	if (CreateHardLinkW(wdest, wsrc, NULL)) {
+		retval = 0;
+		/* debug_print("hard link created: %s -> %s\n", src, dest); */
+	} else {
+		retval = -1;
+		switch (GetLastError()) {
+		case ERROR_FILE_NOT_FOUND:
+		case ERROR_PATH_NOT_FOUND:
+			errno = ENOENT; break;
+		case ERROR_ACCESS_DENIED:
+		case ERROR_LOCK_VIOLATION:
+		case ERROR_SHARING_VIOLATION:
+			errno = EACCES; break;
+		case ERROR_NOT_SAME_DEVICE:
+			errno = EXDEV; break;
+		case ERROR_FILE_EXISTS:
+		case ERROR_ALREADY_EXISTS:
+			errno = EEXIST; break;
+		case ERROR_TOO_MANY_LINKS:
+			errno = EMLINK; break;
+		default:
+			errno = EIO; break;
+		}
+	}
+	save_errno = errno;
+
+	g_free(wdest);
+	g_free(wsrc);
+
+	errno = save_errno;
+	return retval;
+#else
+	return link(src, dest);
+#endif
+}
+#endif /* !G_OS_UNIX */
+
 void list_free_strings(GList *list)
 {
 	list = g_list_first(list);
@@ -252,6 +314,24 @@ gint to_number(const gchar *nstr)
 	return atoi(nstr);
 }
 
+guint to_unumber(const gchar *nstr)
+{
+	register const gchar *p;
+	gulong val;
+
+	if (*nstr == '\0') return 0;
+
+	for (p = nstr; *p != '\0'; p++)
+		if (!g_ascii_isdigit(*p)) return 0;
+
+	errno = 0;
+	val = strtoul(nstr, NULL, 10);
+	if (val == ULONG_MAX && errno != 0)
+		val = 0;
+
+	return (guint)val;
+}
+
 /* convert integer into string,
    nstr must be not lower than 11 characters length */
 gchar *itos_buf(gchar *nstr, gint n)
@@ -268,6 +348,12 @@ gchar *itos(gint n)
 	return itos_buf(nstr, n);
 }
 
+gchar *utos_buf(gchar *nstr, guint n)
+{
+	g_snprintf(nstr, 11, "%u", n);
+	return nstr;
+}
+
 gchar *to_human_readable(gint64 size)
 {
 	static gchar str[16];
@@ -324,6 +410,33 @@ gint path_cmp(const gchar *s1, const gchar *s2)
 #endif
 }
 
+/* return TRUE if parent is equal to or ancestor of child */
+gboolean is_path_parent(const gchar *parent, const gchar *child)
+{
+	gint plen;
+	const gchar *base;
+
+	g_return_val_if_fail(parent != NULL, FALSE);
+	g_return_val_if_fail(child != NULL, FALSE);
+
+	plen = strlen(parent);
+	while (plen > 0 && G_IS_DIR_SEPARATOR(parent[plen - 1]))
+		plen--;
+
+#ifdef G_OS_WIN32
+	if (!g_ascii_strncasecmp(parent, child, plen)) {
+#else
+	if (!strncmp(parent, child, plen)) {
+#endif
+		base = child + plen;
+		if (!G_IS_DIR_SEPARATOR(*base) && *base != '\0')
+			return FALSE;
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
 /* remove trailing return code */
 gchar *strretchomp(gchar *str)
 {
@@ -398,18 +511,21 @@ gpointer my_memmem(gconstpointer haystack, size_t haystacklen,
 	const gchar *haystack_ = (const gchar *)haystack;
 	const gchar *needle_ = (const gchar *)needle;
 	const gchar *haystack_cur = (const gchar *)haystack;
+	size_t haystack_left = haystacklen;
 
 	if (needlelen == 1)
 		return memchr(haystack_, *needle_, haystacklen);
 
-	while ((haystack_cur = memchr(haystack_cur, *needle_, haystacklen))
+	while ((haystack_cur = memchr(haystack_cur, *needle_, haystack_left))
 	       != NULL) {
 		if (haystacklen - (haystack_cur - haystack_) < needlelen)
 			break;
 		if (memcmp(haystack_cur + 1, needle_ + 1, needlelen - 1) == 0)
 			return (gpointer)haystack_cur;
-		else
+		else {
 			haystack_cur++;
+			haystack_left = haystacklen - (haystack_cur - haystack_);
+		}
 	}
 
 	return NULL;
@@ -489,7 +605,6 @@ gboolean is_next_nonascii(const gchar *s)
 gint get_next_word_len(const gchar *s)
 {
 	const gchar *p = s;
-	gint len = 0;
 	gboolean in_quote = FALSE;
 
 	while (*p != '\0') {
@@ -850,6 +965,30 @@ void extract_list_id_str(gchar *str)
 	g_strstrip(str);
 }
 
+gchar *extract_addresses(const gchar *str)
+{
+	GString *new_str;
+	GSList *addr_list, *cur;
+
+	if (!str)
+		return NULL;
+
+	addr_list = address_list_append(NULL, str);
+
+	new_str = g_string_new(NULL);
+
+	for (cur = addr_list; cur != NULL; cur = cur->next) {
+		g_string_append(new_str, (gchar *)cur->data);
+		if (cur->next)
+			g_string_append(new_str, ", ");
+	}
+
+	slist_free_strings(addr_list);
+	g_slist_free(addr_list);
+
+	return g_string_free(new_str, FALSE);
+}
+
 gchar *normalize_address_field(const gchar *str)
 {
 	GString *new_str;
@@ -1326,6 +1465,43 @@ gchar *strstr_with_skip_quote(const gchar *haystack, const gchar *needle)
 	return NULL;
 }
 
+gchar *strcasestr_with_skip_quote(const gchar *haystack, const gchar *needle)
+{
+	register guint haystack_len, needle_len;
+	gboolean in_squote = FALSE, in_dquote = FALSE;
+
+	haystack_len = strlen(haystack);
+	needle_len   = strlen(needle);
+
+	if (haystack_len < needle_len || needle_len == 0)
+		return NULL;
+
+	while (haystack_len >= needle_len) {
+		if (!in_squote && !in_dquote &&
+		    !g_ascii_strncasecmp(haystack, needle, needle_len))
+			return (gchar *)haystack;
+
+		/* 'foo"bar"' -> foo"bar"
+		   "foo'bar'" -> foo'bar' */
+		if (*haystack == '\'') {
+			if (in_squote)
+				in_squote = FALSE;
+			else if (!in_dquote)
+				in_squote = TRUE;
+		} else if (*haystack == '\"') {
+			if (in_dquote)
+				in_dquote = FALSE;
+			else if (!in_squote)
+				in_dquote = TRUE;
+		}
+
+		haystack++;
+		haystack_len--;
+	}
+
+	return NULL;
+}
+
 gchar *strchr_parenthesis_close(const gchar *str, gchar op, gchar cl)
 {
 	const gchar *p;
@@ -2028,6 +2204,15 @@ const gchar *get_document_dir(void)
 			document_dir = get_home_dir();
 	}
 
+	return document_dir;
+#elif defined(__APPLE__)
+	static const gchar *document_dir = NULL;
+
+	if (!document_dir) {
+		document_dir = g_strconcat(get_home_dir(), G_DIR_SEPARATOR_S,
+					   "Documents", NULL);
+	}
+
 	return document_dir;
 #else
 	return get_home_dir();
@@ -2048,6 +2233,11 @@ const gchar *get_rc_dir(void)
 			rc_dir = g_strconcat(get_home_dir(), G_DIR_SEPARATOR_S,
 					     RC_DIR, NULL);
 		g_free(appdata);
+#elif defined(__APPLE__)
+		rc_dir = g_strconcat(get_home_dir(), G_DIR_SEPARATOR_S,
+				     "Library", G_DIR_SEPARATOR_S,
+				     "Application Support", G_DIR_SEPARATOR_S,
+				     RC_DIR, NULL);
 #else
 		rc_dir = g_strconcat(get_home_dir(), G_DIR_SEPARATOR_S,
 				     RC_DIR, NULL);
@@ -2070,7 +2260,7 @@ const gchar *get_old_rc_dir(void)
 
 const gchar *get_mail_base_dir(void)
 {
-#ifdef G_OS_WIN32
+#if defined(G_OS_WIN32) || defined(__APPLE__)
 	static gchar *mail_base_dir = NULL;
 
 	if (!mail_base_dir)
@@ -2246,6 +2436,49 @@ off_t get_left_file_size(FILE *fp)
 	return size;
 }
 
+gint get_last_empty_line_size(FILE *fp, off_t size)
+{
+	glong pos;
+	gint lsize = 0;
+	gchar buf[4];
+	size_t nread;
+
+	if (size < 4)
+		return -1;
+
+	if ((pos = ftell(fp)) < 0) {
+		perror("ftell");
+		return -1;
+	}
+	if (fseek(fp, size - 4, SEEK_CUR) < 0) {
+		perror("fseek");
+		return -1;
+	}
+
+	/* read last 4 bytes */
+	nread = fread(buf, sizeof(buf), 1, fp);
+	if (nread != 1) {
+		perror("fread");
+		return -1;
+	}
+	/* g_print("last 4 bytes: %02x %02x %02x %02x\n", buf[0], buf[1], buf[2], buf[3]); */
+	if (buf[3] == '\n') {
+		if (buf[2] == '\n')
+			lsize = 1;
+		else if (buf[2] == '\r') {
+			if (buf[1] == '\n')
+				lsize = 2;
+		}
+	}
+
+	if (fseek(fp, pos, SEEK_SET) < 0) {
+		perror("fseek");
+		return -1;
+	}
+
+	return lsize;
+}
+
 gboolean file_exist(const gchar *file, gboolean allow_fifo)
 {
 	if (file == NULL)
@@ -2412,7 +2645,7 @@ gint remove_numbered_files(const gchar *dir, guint first, guint last)
 	GDir *dp;
 	const gchar *dir_name;
 	gchar *prev_dir;
-	gint file_no;
+	guint file_no;
 
 	prev_dir = g_get_current_dir();
 
@@ -2429,7 +2662,7 @@ gint remove_numbered_files(const gchar *dir, guint first, guint last)
 	}
 
 	while ((dir_name = g_dir_read_name(dp)) != NULL) {
-		file_no = to_number(dir_name);
+		file_no = to_unumber(dir_name);
 		if (file_no > 0 && first <= file_no && file_no <= last) {
 			if (is_dir_exist(dir_name))
 				continue;
@@ -2462,7 +2695,7 @@ gint remove_expired_files(const gchar *dir, guint hours)
 	const gchar *dir_name;
 	struct stat s;
 	gchar *prev_dir;
-	gint file_no;
+	guint file_no;
 	time_t mtime, now, expire_time;
 
 	prev_dir = g_get_current_dir();
@@ -2483,7 +2716,7 @@ gint remove_expired_files(const gchar *dir, guint hours)
 	expire_time = hours * 60 * 60;
 
 	while ((dir_name = g_dir_read_name(dp)) != NULL) {
-		file_no = to_number(dir_name);
+		file_no = to_unumber(dir_name);
 		if (file_no > 0) {
 			if (g_stat(dir_name, &s) < 0) {
 				FILE_OP_ERROR(dir_name, "stat");
@@ -2595,6 +2828,8 @@ gint remove_dir_recursive(const gchar *dir)
 	gchar *cur_dir;
 	gint ret;
 
+	debug_print("remove_dir_recursive: %s\n", dir);
+
 	cur_dir = g_get_current_dir();
 
 	if (g_chdir(dir) < 0) {
@@ -2639,17 +2874,54 @@ gint rename_force(const gchar *oldpath, const gchar *newpath)
 
 gint copy_file(const gchar *src, const gchar *dest, gboolean keep_backup)
 {
+#ifdef G_OS_WIN32
+	wchar_t *wsrc;
+	wchar_t *wdest;
+	gchar *dest_bak = NULL;
+	gboolean err = FALSE;
+
+	wsrc = g_utf8_to_utf16(src, -1, NULL, NULL, NULL);
+	if (wsrc == NULL) {
+		return -1;
+	}
+	wdest = g_utf8_to_utf16(dest, -1, NULL, NULL, NULL);
+	if (wdest == NULL) {
+		g_free(wsrc);
+		return -1;
+	}
+
+	if (keep_backup == FALSE) {
+		if (CopyFileW(wsrc, wdest, FALSE) == 0)
+			err = TRUE;
+		g_free(wdest);
+		g_free(wsrc);
+		return err ? -1 : 0;
+	}
+
+	if (is_file_exist(dest)) {
+		dest_bak = g_strconcat(dest, ".bak", NULL);
+		if (rename_force(dest, dest_bak) < 0) {
+			FILE_OP_ERROR(dest, "rename");
+			g_free(dest_bak);
+			g_free(wdest);
+			g_free(wsrc);
+			return -1;
+		}
+	}
+
+	if (CopyFileW(wsrc, wdest, FALSE) == 0)
+		err = TRUE;
+
+	g_free(wdest);
+	g_free(wsrc);
+#else
 	gint srcfd, destfd;
 	gint n_read;
 	gchar buf[BUFFSIZE];
 	gchar *dest_bak = NULL;
 	gboolean err = FALSE;
 
-#ifdef G_OS_WIN32
-	if ((srcfd = g_open(src, O_RDONLY | O_BINARY, 0600)) < 0) {
-#else
 	if ((srcfd = g_open(src, O_RDONLY, 0600)) < 0) {
-#endif
 		FILE_OP_ERROR(src, "open");
 		return -1;
 	}
@@ -2663,11 +2935,7 @@ gint copy_file(const gchar *src, const gchar *dest, gboolean keep_backup)
 		}
 	}
 
-#ifdef G_OS_WIN32
-	if ((destfd = g_open(dest, O_WRONLY | O_CREAT | O_BINARY, 0600)) < 0) {
-#else
 	if ((destfd = g_open(dest, O_WRONLY | O_CREAT, 0600)) < 0) {
-#endif
 		FILE_OP_ERROR(dest, "open");
 		close(srcfd);
 		if (dest_bak) {
@@ -2705,6 +2973,7 @@ gint copy_file(const gchar *src, const gchar *dest, gboolean keep_backup)
 		err = TRUE;
 	}
 	close(srcfd);
+#endif
 
 	if (err) {
 		g_unlink(dest);
@@ -2777,19 +3046,52 @@ gint move_file(const gchar *src, const gchar *dest, gboolean overwrite)
 	return 0;
 }
 
-gint copy_file_part(FILE *fp, off_t offset, size_t length, const gchar *dest)
+gint append_file_part(FILE *fp, off_t offset, size_t length, FILE *dest_fp)
 {
-	FILE *dest_fp;
 	gint n_read;
 	gint bytes_left, to_read;
 	gchar buf[BUFSIZ];
-	gboolean err = FALSE;
+
+	g_return_val_if_fail(fp != NULL, -1);
+	g_return_val_if_fail(dest_fp != NULL, -1);
 
 	if (fseek(fp, offset, SEEK_SET) < 0) {
 		perror("fseek");
 		return -1;
 	}
 
+	bytes_left = length;
+	to_read = MIN(bytes_left, sizeof(buf));
+
+	while ((n_read = fread(buf, sizeof(gchar), to_read, fp)) > 0) {
+		if (n_read < to_read && ferror(fp))
+			break;
+		if (fwrite(buf, n_read, 1, dest_fp) < 1) {
+			g_warning("append_file_part: writing to file failed.\n");
+			return -1;
+		}
+		bytes_left -= n_read;
+		if (bytes_left == 0)
+			break;
+		to_read = MIN(bytes_left, sizeof(buf));
+	}
+
+	if (ferror(fp)) {
+		perror("fread");
+		return -1;
+	}
+	if (fflush(dest_fp) == EOF) {
+		FILE_OP_ERROR("append_file_part", "fflush");
+		return -1;
+	}
+
+	return 0;
+}
+
+gint copy_file_part(FILE *fp, off_t offset, size_t length, const gchar *dest)
+{
+	FILE *dest_fp;
+
 	if ((dest_fp = g_fopen(dest, "wb")) == NULL) {
 		FILE_OP_ERROR(dest, "fopen");
 		return -1;
@@ -2800,34 +3102,15 @@ gint copy_file_part(FILE *fp, off_t offset, size_t length, const gchar *dest)
 		g_warning("can't change file mode\n");
 	}
 
-	bytes_left = length;
-	to_read = MIN(bytes_left, sizeof(buf));
-
-	while ((n_read = fread(buf, sizeof(gchar), to_read, fp)) > 0) {
-		if (n_read < to_read && ferror(fp))
-			break;
-		if (fwrite(buf, n_read, 1, dest_fp) < 1) {
-			g_warning(_("writing to %s failed.\n"), dest);
-			fclose(dest_fp);
-			g_unlink(dest);
-			return -1;
-		}
-		bytes_left -= n_read;
-		if (bytes_left == 0)
-			break;
-		to_read = MIN(bytes_left, sizeof(buf));
+	if (append_file_part(fp, offset, length, dest_fp) < 0) {
+		g_warning("writing to %s failed.\n", dest);
+		fclose(dest_fp);
+		g_unlink(dest);
+		return -1;
 	}
 
-	if (ferror(fp)) {
-		perror("fread");
-		err = TRUE;
-	}
 	if (fclose(dest_fp) == EOF) {
 		FILE_OP_ERROR(dest, "fclose");
-		err = TRUE;
-	}
-
-	if (err) {
 		g_unlink(dest);
 		return -1;
 	}
@@ -3764,11 +4047,73 @@ gint execute_command_line(const gchar *cmdline, gboolean async)
 	return ret;
 }
 
+#if USE_THREADS
+typedef struct _CmdData
+{
+	const gchar *cmdline;
+	volatile gint flag;
+	gint status;
+} CmdData;
+
+static gpointer execute_command_line_async_func(gpointer data)
+{
+	CmdData *cmd_data = (CmdData *)data;
+	gchar **argv;
+
+	argv = strsplit_with_quote(cmd_data->cmdline, " ", 0);
+	cmd_data->status = execute_sync(argv);
+	g_strfreev(argv);
+
+	debug_print("execute_command_line_async_func: exec done: %s\n",
+		    cmd_data->cmdline);
+	g_atomic_int_set(&cmd_data->flag, 1);
+	g_main_context_wakeup(NULL);
+
+	return GINT_TO_POINTER(0);
+}
+
+gint execute_command_line_async_wait(const gchar *cmdline)
+{
+	CmdData data = {NULL, 0, 0};
+	GThread *thread;
+
+	if (debug_mode) {
+		gchar *utf8_cmdline;
+
+		utf8_cmdline = g_filename_to_utf8
+			(cmdline, -1, NULL, NULL, NULL);
+		debug_print("execute_command_line(): executing: %s\n",
+			    utf8_cmdline ? utf8_cmdline : cmdline);
+		g_free(utf8_cmdline);
+	}
+
+	data.cmdline = cmdline;
+	thread = g_thread_create(execute_command_line_async_func, &data, TRUE,
+				 NULL);
+	if (!thread)
+		return -1;
+
+	debug_print("execute_command_line_async_wait: waiting thread\n");
+	while (g_atomic_int_get(&data.flag) == 0)
+		event_loop_iterate();
+
+	g_thread_join(thread);
+	debug_print("execute_command_line_async_wait: thread exited\n");
+
+	return data.status;
+}
+#else /* USE_THREADS */
+gint execute_command_line_async_wait(const gchar *cmdline)
+{
+	return execute_command_line(cmdline, FALSE);
+}
+#endif /* USE_THREADS */
+
 gint execute_open_file(const gchar *file, const gchar *content_type)
 {
+#ifdef G_OS_WIN32
 	g_return_val_if_fail(file != NULL, -1);
 
-#ifdef G_OS_WIN32
 	log_print("opening %s - %s\n", file, content_type ? content_type : "");
 
 	if (G_WIN32_HAVE_WIDECHAR_API()) {
@@ -3796,7 +4141,17 @@ gint execute_open_file(const gchar *file, const gchar *content_type)
 
 		return 0;
 	}
+#elif defined(__APPLE__)
+	const gchar *argv[3] = {"open", NULL, NULL};
+
+	g_return_val_if_fail(file != NULL, -1);
+
+	log_print("opening %s - %s\n", file, content_type ? content_type : "");
+
+	argv[1] = file;
+	execute_async(argv);
 #endif
+
 	return 0;
 }
 
@@ -3861,7 +4216,7 @@ gint open_uri(const gchar *uri, const gchar *cmdline)
 
 	g_return_val_if_fail(uri != NULL, -1);
 
-#ifdef G_OS_WIN32
+#if defined(G_OS_WIN32) || defined(__APPLE__)
 	if (!cmdline || cmdline[0] == '\0')
 		return execute_open_file(uri, NULL);
 #endif
@@ -3979,10 +4334,9 @@ time_t tzoffset_sec(time_t *now)
 	return off * 60;
 }
 
-/* calculate timezone offset */
-gchar *tzoffset(time_t *now)
+/* calculate timezone offset (buf must not be less than 6 bytes) */
+gchar *tzoffset_buf(gchar *buf, time_t *now)
 {
-	static gchar offset_string[6];
 	struct tm gmt, *tmp, *lt;
 	gint off;
 	gchar sign = '+';
@@ -4012,9 +4366,16 @@ gchar *tzoffset(time_t *now)
 	if (off >= 24 * 60)		/* should be impossible */
 		off = 23 * 60 + 59;	/* if not, insert silly value */
 
-	sprintf(offset_string, "%c%02d%02d", sign, off / 60, off % 60);
+	g_snprintf(buf, 6, "%c%02d%02d", sign, off / 60, off % 60);
 
-	return offset_string;
+	return buf;
+}
+
+gchar *tzoffset(time_t *now)
+{
+	static gchar offset_string[6];
+
+	return tzoffset_buf(offset_string, now);
 }
 
 void get_rfc822_date(gchar *buf, gint len)
@@ -4023,6 +4384,7 @@ void get_rfc822_date(gchar *buf, gint len)
 	time_t t;
 	gchar day[4], mon[4];
 	gint dd, hh, mm, ss, yyyy;
+	gchar off[6];
 
 	t = time(NULL);
 	lt = localtime(&t);
@@ -4030,7 +4392,7 @@ void get_rfc822_date(gchar *buf, gint len)
 	sscanf(asctime(lt), "%3s %3s %d %d:%d:%d %d\n",
 	       day, mon, &dd, &hh, &mm, &ss, &yyyy);
 	g_snprintf(buf, len, "%s, %d %s %d %02d:%02d:%02d %s",
-		   day, dd, mon, yyyy, hh, mm, ss, tzoffset(&t));
+		   day, dd, mon, yyyy, hh, mm, ss, tzoffset_buf(off, &t));
 }
 
 /* just a wrapper to suppress the warning of gcc about %c */
@@ -4055,6 +4417,21 @@ void ui_update(void)
 		ui_update_func();
 }
 
+static EventLoopFunc event_loop_func = NULL;
+
+void set_event_loop_func(EventLoopFunc func)
+{
+	event_loop_func = func;
+}
+
+void event_loop_iterate(void)
+{
+	if (event_loop_func)
+		event_loop_func();
+	else
+		g_main_context_iteration(NULL, TRUE);
+}
+
 static ProgressFunc progress_func = NULL;
 
 void set_progress_func(ProgressFunc func)
@@ -4088,21 +4465,34 @@ gchar *input_query_password(const gchar *server, const gchar *user)
 /* logging */
 
 static FILE *log_fp = NULL;
+#if USE_THREADS
+G_LOCK_DEFINE_STATIC(log_fp);
+#define S_LOCK(name)	G_LOCK(name)
+#define S_UNLOCK(name)	G_UNLOCK(name)
+#else
+#define S_LOCK(name)
+#define S_UNLOCK(name)
+#endif
 
 void set_log_file(const gchar *filename)
 {
-	if (log_fp) return;
-	log_fp = g_fopen(filename, "wb");
-	if (!log_fp)
-		FILE_OP_ERROR(filename, "fopen");
+	S_LOCK(log_fp);
+	if (!log_fp) {
+		log_fp = g_fopen(filename, "w");
+		if (!log_fp)
+			FILE_OP_ERROR(filename, "fopen");
+	}
+	S_UNLOCK(log_fp);
 }
 
 void close_log_file(void)
 {
+	S_LOCK(log_fp);
 	if (log_fp) {
 		fclose(log_fp);
 		log_fp = NULL;
 	}
+	S_UNLOCK(log_fp);
 }
 
 static guint log_verbosity_count = 0;
@@ -4129,10 +4519,15 @@ static void log_dummy_func(const gchar *str)
 {
 }
 
+static void log_dummy_flush_func(void)
+{
+}
+
 static LogFunc log_print_ui_func = log_dummy_func;
 static LogFunc log_message_ui_func = log_dummy_func;
 static LogFunc log_warning_ui_func = log_dummy_func;
 static LogFunc log_error_ui_func = log_dummy_func;
+static LogFlushFunc log_flush_ui_func = log_dummy_flush_func;
 
 static LogFunc log_show_status_func = log_dummy_func;
 
@@ -4145,6 +4540,14 @@ void set_log_ui_func(LogFunc print_func, LogFunc message_func,
 	log_error_ui_func = error_func;
 }
 
+void set_log_ui_func_full(LogFunc print_func, LogFunc message_func,
+			  LogFunc warning_func, LogFunc error_func,
+			  LogFlushFunc flush_func)
+{
+	set_log_ui_func(print_func, message_func, warning_func, error_func);
+	log_flush_ui_func = flush_func;
+}
+
 void set_log_show_status_func(LogFunc status_func)
 {
 	log_show_status_func = status_func;
@@ -4180,6 +4583,8 @@ void status_print(const gchar *format, ...)
 
 void log_write(const gchar *str, const gchar *prefix)
 {
+	S_LOCK(log_fp);
+
 	if (log_fp) {
 		gchar buf[TIME_LEN + 1];
 		time_t t;
@@ -4193,6 +4598,8 @@ void log_write(const gchar *str, const gchar *prefix)
 		fputs(str, log_fp);
 		fflush(log_fp);
 	}
+
+	S_UNLOCK(log_fp);
 }
 
 void log_print(const gchar *format, ...)
@@ -4210,10 +4617,12 @@ void log_print(const gchar *format, ...)
 
 	if (debug_mode) g_print("%s", buf);
 	log_print_ui_func(buf);
+	S_LOCK(log_fp);
 	if (log_fp) {
 		fputs(buf, log_fp);
 		fflush(log_fp);
 	}
+	S_UNLOCK(log_fp);
 	if (log_verbosity_count)
 		log_show_status_func(buf + TIME_LEN);
 }
@@ -4233,12 +4642,14 @@ void log_message(const gchar *format, ...)
 
 	if (debug_mode) g_message("%s", buf + TIME_LEN);
 	log_message_ui_func(buf + TIME_LEN);
+	S_LOCK(log_fp);
 	if (log_fp) {
 		fwrite(buf, TIME_LEN, 1, log_fp);
 		fputs("* message: ", log_fp);
 		fputs(buf + TIME_LEN, log_fp);
 		fflush(log_fp);
 	}
+	S_UNLOCK(log_fp);
 	log_show_status_func(buf + TIME_LEN);
 }
 
@@ -4257,12 +4668,14 @@ void log_warning(const gchar *format, ...)
 
 	g_warning("%s", buf);
 	log_warning_ui_func(buf + TIME_LEN);
+	S_LOCK(log_fp);
 	if (log_fp) {
 		fwrite(buf, TIME_LEN, 1, log_fp);
 		fputs("** warning: ", log_fp);
 		fputs(buf + TIME_LEN, log_fp);
 		fflush(log_fp);
 	}
+	S_UNLOCK(log_fp);
 }
 
 void log_error(const gchar *format, ...)
@@ -4280,10 +4693,21 @@ void log_error(const gchar *format, ...)
 
 	g_warning("%s", buf);
 	log_error_ui_func(buf + TIME_LEN);
+	S_LOCK(log_fp);
 	if (log_fp) {
 		fwrite(buf, TIME_LEN, 1, log_fp);
 		fputs("*** error: ", log_fp);
 		fputs(buf + TIME_LEN, log_fp);
 		fflush(log_fp);
 	}
+	S_UNLOCK(log_fp);
+}
+
+void log_flush(void)
+{
+	S_LOCK(log_fp);
+	if (log_fp)
+		fflush(log_fp);
+	S_UNLOCK(log_fp);
+	log_flush_ui_func();
 }
