@@ -206,7 +206,7 @@ MimeView *mimeview_create(void)
 	renderer = gtk_cell_renderer_text_new();
 	g_object_set(renderer, "ypad", 0, NULL);
 	column = gtk_tree_view_column_new_with_attributes
-		(_("MIME Type"), renderer, "text", COL_MIMETYPE, NULL);
+		(_("Data type"), renderer, "text", COL_MIMETYPE, NULL);
 	gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
 	gtk_tree_view_column_set_resizable(column, TRUE);
 	gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
@@ -821,8 +821,11 @@ static void mimeview_selection_changed(GtkTreeSelection *selection,
 		mimeview_show_message_part(mimeview, partinfo);
 		break;
 	case MIME_IMAGE:
-		mimeview_show_image_part(mimeview, partinfo);
-		break;
+		if (prefs_common.inline_image) {
+			mimeview_show_image_part(mimeview, partinfo);
+			break;
+		}
+		/* FALLTHROUGH */
 	default:
 		mimeview_change_view_type(mimeview, MIMEVIEW_TEXT);
 #if USE_GPGME
@@ -1223,6 +1226,21 @@ static void mimeview_view_file(const gchar *filename, MimeInfo *partinfo,
 		}
 		execute_open_file(filename, partinfo->content_type);
 		return;
+#elif defined(__APPLE__)
+		if (g_file_test(filename, G_FILE_TEST_IS_EXECUTABLE) ||
+		    str_has_suffix_case(filename, ".py") ||
+		    str_has_suffix_case(filename, ".rb") ||
+		    str_has_suffix_case(filename, ".sh")) {
+			alertpanel_full
+				(_("Opening executable file"),
+				 _("This is an executable file. Opening executable file is restricted for security.\n"
+				   "If you want to launch it, save it to somewhere and make sure it is not an virus or something like a malicious program."),
+				 ALERT_WARNING, G_ALERTDEFAULT, FALSE,
+				 GTK_STOCK_OK, NULL, NULL);
+			return;
+		}
+		execute_open_file(filename, partinfo->content_type);
+		return;
 #else
 		if (MIME_IMAGE == partinfo->mime_type)
 			cmd = prefs_common.mime_image_viewer;
