@@ -1,6 +1,6 @@
 /*
  * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
- * Copyright (C) 1999-2005 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2008 Hiroyuki Yamamoto
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -34,6 +34,7 @@
 
 #include "main.h"
 #include "prefs.h"
+#include "prefs_ui.h"
 #include "prefs_filter.h"
 #include "prefs_filter_edit.h"
 #include "prefs_common.h"
@@ -85,10 +86,6 @@ static void prefs_filter_set_list_row		(GtkTreeIter	*iter,
 						 FilterRule	*rule,
 						 gboolean	 move_view);
 
-static void prefs_filter_set_header_list	(MsgInfo	*msginfo);
-
-static void prefs_filter_write_user_header_list	(void);
-
 static void prefs_filter_set_list		(void);
 
 /* callback functions */
@@ -129,7 +126,7 @@ static gboolean prefs_filter_key_pressed(GtkWidget	*widget,
 static void prefs_filter_close		(void);
 
 
-void prefs_filter_open(MsgInfo *msginfo, const gchar *header)
+void prefs_filter_open(MsgInfo *msginfo, const gchar *header, const gchar *key)
 {
 	inc_lock();
 
@@ -144,11 +141,13 @@ void prefs_filter_open(MsgInfo *msginfo, const gchar *header)
 	prefs_filter_set_dialog();
 
 	gtk_widget_show(rule_list_window.window);
+	manage_window_focus_in(rule_list_window.window, NULL, NULL);
 
 	if (msginfo) {
 		FilterRule *rule;
 
-		rule = prefs_filter_edit_open(NULL, header);
+		rule = prefs_filter_edit_open(NULL, header, key);
+		gtk_window_present(GTK_WINDOW(rule_list_window.window));
 
 		if (rule) {
 			prefs_filter_set_list_row(NULL, rule, TRUE);
@@ -206,7 +205,7 @@ static void prefs_filter_create(void)
 	gtk_widget_grab_default(close_btn);
 
 	gtk_window_set_title(GTK_WINDOW(window),
-			     _("Filter setting"));
+			     _("Filter settings"));
 	g_signal_connect(G_OBJECT(window), "delete_event",
 			 G_CALLBACK(prefs_filter_deleted), NULL);
 	g_signal_connect(G_OBJECT(window), "key_press_event",
@@ -359,84 +358,6 @@ static void prefs_filter_create(void)
 	rule_list_window.msg_hdr_table = NULL;
 }
 
-void prefs_filter_read_config(void)
-{
-	gchar *rcpath;
-	GNode *node;
-	FilterRule *rule;
-
-	debug_print("Reading filter configuration...\n");
-
-	/* remove all previous filter list */
-	while (prefs_common.fltlist != NULL) {
-		rule = (FilterRule *)prefs_common.fltlist->data;
-		filter_rule_free(rule);
-		prefs_common.fltlist = g_slist_remove(prefs_common.fltlist,
-						      rule);
-	}
-
-	rcpath = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S, FILTER_LIST,
-			     NULL);
-	if (!is_file_exist(rcpath)) {
-		g_free(rcpath);
-		return;
-	}
-
-	node = xml_parse_file(rcpath);
-	if (!node) {
-		g_warning("Can't parse %s\n", rcpath);
-		g_free(rcpath);
-		return;
-	}
-	g_free(rcpath);
-
-	prefs_common.fltlist = filter_xml_node_to_filter_list(node);
-
-	xml_free_tree(node);
-}
-
-void prefs_filter_write_config(void)
-{
-	filter_write_config(prefs_common.fltlist);
-}
-
-void prefs_filter_rename_path(const gchar *old_path, const gchar *new_path)
-{
-	GSList *cur;
-
-	g_return_if_fail(old_path != NULL);
-	g_return_if_fail(new_path != NULL);
-
-	for (cur = prefs_common.fltlist; cur != NULL; cur = cur->next) {
-		FilterRule *rule = (FilterRule *)cur->data;
-		filter_rule_rename_dest_path(rule, old_path, new_path);
-	}
-
-	filter_write_config(prefs_common.fltlist);
-}
-
-void prefs_filter_delete_path(const gchar *path)
-{
-	GSList *cur;
-	GSList *next;
-
-	g_return_if_fail(path != NULL);
-
-	for (cur = prefs_common.fltlist; cur != NULL; cur = next) {
-		FilterRule *rule = (FilterRule *)cur->data;
-		next = cur->next;
-
-		filter_rule_delete_action_by_dest_path(rule, path);
-		if (!rule->action_list) {
-			prefs_common.fltlist =
-				g_slist_remove(prefs_common.fltlist, rule);
-			filter_rule_free(rule);
-		}
-	}
-
-	filter_write_config(prefs_common.fltlist);
-}
-
 static void prefs_filter_set_dialog(void)
 {
 	GSList *cur;
@@ -508,7 +429,8 @@ static void prefs_filter_set_list_row(GtkTreeIter *iter, FilterRule *rule,
 									  \
 		if (!g_hash_table_lookup(table, header->name)) {	  \
 			g_hash_table_insert(table, header->name, header); \
-			list = g_slist_append(list, header);		  \
+			list = procheader_add_header_list		  \
+				(list, header->name, header->body);	  \
 		}							  \
 	}
 
@@ -603,7 +525,7 @@ void prefs_filter_set_msg_header_list(MsgInfo *msginfo)
 	}
 }
 
-static void prefs_filter_set_header_list(MsgInfo *msginfo)
+void prefs_filter_set_header_list(MsgInfo *msginfo)
 {
 	GSList *list = NULL;
 	gchar *path;
@@ -623,7 +545,7 @@ static void prefs_filter_set_header_list(MsgInfo *msginfo)
 	list = NULL;
 	path = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S, FILTER_HEADER_RC,
 			   NULL);
-	if ((fp = fopen(path, "rb")) != NULL) {
+	if ((fp = g_fopen(path, "rb")) != NULL) {
 		gchar buf[PREFSBUFSIZE];
 
 		while (fgets(buf, sizeof(buf), fp) != NULL) {
@@ -638,11 +560,10 @@ static void prefs_filter_set_header_list(MsgInfo *msginfo)
 	g_free(path);
 
 	prefs_filter_set_user_header_list(list);
-
 	prefs_filter_set_msg_header_list(msginfo);
 }
 
-static void prefs_filter_write_user_header_list(void)
+void prefs_filter_write_user_header_list(void)
 {
 	gchar *path;
 	PrefFile *pfile;
@@ -693,7 +614,8 @@ static void prefs_filter_add_cb(void)
 {
 	FilterRule *rule;
 
-	rule = prefs_filter_edit_open(NULL, NULL);
+	rule = prefs_filter_edit_open(NULL, NULL, NULL);
+	gtk_window_present(GTK_WINDOW(rule_list_window.window));
 
 	if (rule) {
 		prefs_filter_set_list_row(NULL, rule, TRUE);
@@ -714,7 +636,8 @@ static void prefs_filter_edit_cb(void)
 			   COL_FILTER_RULE, &rule, -1);
 	g_return_if_fail(rule != NULL);
 
-	new_rule = prefs_filter_edit_open(rule, NULL);
+	new_rule = prefs_filter_edit_open(rule, NULL, NULL);
+	gtk_window_present(GTK_WINDOW(rule_list_window.window));
 
 	if (new_rule) {
 		prefs_filter_set_list_row(&iter, new_rule, TRUE);
@@ -735,7 +658,8 @@ static void prefs_filter_copy_cb(void)
 			   COL_FILTER_RULE, &rule, -1);
 	g_return_if_fail(rule != NULL);
 
-	new_rule = prefs_filter_edit_open(rule, NULL);
+	new_rule = prefs_filter_edit_open(rule, NULL, NULL);
+	gtk_window_present(GTK_WINDOW(rule_list_window.window));
 
 	if (new_rule) {
 		prefs_filter_set_list_row(NULL, new_rule, TRUE);
@@ -897,10 +821,10 @@ static gboolean prefs_filter_key_pressed(GtkWidget *widget, GdkEventKey *event,
 static void prefs_filter_close(void)
 {
 	prefs_filter_set_msg_header_list(NULL);
-	prefs_filter_write_user_header_list();
 	prefs_filter_set_list();
-	filter_write_config(prefs_common.fltlist);
+	filter_write_config();
 	gtk_widget_hide(rule_list_window.window);
 	gtk_list_store_clear(rule_list_window.store);
+	main_window_popup(main_window_get());
 	inc_unlock();
 }
